;;; Kernel.inc -- definitions for a FORTH-like, Threaded Language ;;;;;;;;;;;;

;
;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Some values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TRUE		=	-1	; Logical True value
FALSE		=	 0	; Logical False value

BYTEBITS	=	 8	; bits in a byte (smallest addressable value)

CHARABITS	=	 0	; bits of char in address
CELLABITS	=	 2	; bits of cell in address
PAGEABITS	=	12	; bits of page in address

VERSION_MAJOR		=	0
VERSION_MINOR		=	0
VERSION_PATCH_LEVEL	=	0
VERSION_INTERNAL	=	3


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Deduced values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CHARSIZE	=	(1 SHL CHARABITS)	; size of a char in bytes
CELLSIZE	=	(1 SHL CELLABITS)	; size of a cell in bytes
PAGESIZE	=	(1 SHL PAGEABITS)	; size of a page in bytes
CHARMASK	=	NOT (CHARSIZE - 1)	; mask of chars in address
CELLMASK	=	NOT (CELLSIZE - 1)	; mask of cells in address
PAGEMASK	=	NOT (PAGESIZE - 1)	; mask of pages in address
BITSPERCHAR	=	CHARSIZE*BYTEBITS	; bits in a char
BITSPERCELL	=	CELLSIZE*BYTEBITS	; bits in a cell


;;;;;;;;;;;;;;;;;;;;;;;;; IMPLEMENTATIONS OF "NEXT" ;;;;;;;;;;;;;;;;;;;;;;;;;;

PC		equ	esp	; program counter
DSP		equ	edi	; data stack pointer
DSINC		=	+4	; increment for data stack
RSP		equ	ebp	; RSP is the return stack pointer (growing up)
RSINC		=	-4	; increment for return stack
TOS		equ	ebx	; buffer for the top of stack
R0		equ	eax	; a scratch register
R0L		equ	al	; its lowest byte
R0H		equ	ah	; its second byte
INSTRTBL	equ	InstrT	; table for bytecode threading


;;; RET-based direct threading
;;; --------------------------
Macro	R_Enter
EndM	R_Enter
Macro	R_Next
	ret		; 1,  10+m, 5
EndM	R_Next


;;; BYTECODE, INDIRECT THREADED
;;; ---------------------------
Macro	BCIT_Enter
EndM	BCIT_Enter
Macro	BCIT_Next
; R0 is callee-saved InstrTbl/4 on high bytes...
; InstrTbl is thus 64-bytes-aligned...
	mov R0L,[PC]		; 2, 4  , 1
	inc PC			; 1, 2  , 1
	jmp [R0*4]		; 3,10+m, 5
EndM	BCIT_Next
	;TOTAL:			; 6,16+m, 7		; or


; External loop for bytecode with rets:
; BCLOOP:
;    push offset BCLOOP
;    mov R0,InstrTbl/4
;    BCIT_Next
; or more efficiently, prepend the next code to the whole routine !


;;;;;;;;;;;;;;;;;;;;;;; EQUATES FOR DEFAULT BEHAVIOUR ;;;;;;;;;;;;;;;;;;;;;;;;;
Threading	equ	R	; Ret based threading

;;; Generic code
;;; ------------
Macro	_Enter
  % Threading&<_Enter>
EndM	_Enter
Macro	Next
  % Threading&<_Next>
EndM	Next




;;;;;;;;;;;;;;;;;;;;;;;;;; Defining FORTH words ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Macro	DeclName	name
	local	len
  Segment NameSpace
    NAMEADDR = $
	db	len
	db	<name>
    len = $ - NAMEADDR - 1
  EndS	NameSpace
EndM	DeclName
Macro	Define		name
	local OBJADDR
	OBJADDR = $
	DeclName	<name>
  Segment NameTable
	dd	NAMEADDR, OBJADDR
  EndS	  NameTable
EndM	Define
Macro	Definline	name	; inline function 
EndM	Definline



;;;;;;;;;;;;;;;;;;;;;;;;;;;; Absolute jump macros ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Macro AbsJmp dest,tmpreg
	ifb <tmpreg>
	  push (offset dest)		; 5,  2  , 1
	  retn				; 1, 10+m, 5
	elif <>
	  push (offset dest)
	  retn
	else
	  mov tmpreg,(offset dest)	; 5,  2  , 1
	  jmp tmpreg			; 2,  7+m, 5
	endif
EndM  AbsJmp
Macro AbsLngJmp dest
	push (seg dest)			; 5
	push (offset dest)		; 5
	retf				; 1
EndM  AbsLngJmp
Macro	ACall dest,align?
	ifnb <align?>
	  AlignP 4,2
	endif
	ds:
	mov eax,offset dest
	call eax
EndM	ACall




;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Virtual arrays ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; parameter is the address of a character*attribute virtual array.

Struc VirtualArray
param	dd	?	; parameter		( va )
width	dd	?	; get number of columns	( va -- n )
height	dd	?	; get number of lines	( va -- n )
get	dd	?	; get char & attribute  ( va x y -- ch attr )
put	dd	?	; set char & attribute  ( va x y ch attr -- )
EndS  VirtualArray


;;;;;;;;;;;;;;;;;;;;;;; Aerial (extra-light) threads ;;;;;;;;;;;;;;;;;;;;;;;;;
Struc AerialThread	_bx,_si,_di,_bp,_sp,eesp,resp ; Lighter than Light !
  AT_SAVE_EBX		dd	?	; initial Top of Stack
  AT_SAVE_ESI		dd	?	; Data Stack Pointer
  AT_SAVE_EDI		dd	?	; Return Stack Pointer
  AT_SAVE_EBP		dd	?	; Those two are reserved for the
  AT_SAVE_ESP		dd	?	;; threading mechanism.
  AT_SAVE_RESP		dd	?	; ESP save when going ret-threaded
  AT_STATE		dd	?	; flags used for thread sync'
  AT_MAX_ESI		dd	?	; bottom of data stack
  AT_MAX_EDI		dd	?	; bottom of return stack
  AT_LINKBACK		dd	?	; link back to thread object
EndS  AerialThread

Macro	END_RETHREAD
	xchg ebp,[Struc AerialThread ebp.AT_SAVE_RESP]
	xchg ebp,esp
EndM	END_RETHREAD
Macro	USE_RETHREAD
	xchg ebp,esp
	xchg ep,[Struc AerialThread ebp.AT_SAVE_RESP]
EndM	USE_RETHREAD

;END_RETHREAD: (esp,resp) -> (ebp,esp)
; old ebp is saved on [esp]